@@ -1,14 +1,18 @@
 # DB Classes
+from __future__ import with_statement
 import sqlalchemy as sql
 import elixir as elixir
 import migrate as migrate
 import database
-from datetime import datetime
 import os, sys, time
 from globals import *
 from urllib import urlopen
 import urlparse
 
+
+# Mark Pilgrim's Feed Parser
+import feedparser as fp
+fp.USER_AGENT = 'uRSSus/%s +http://urssus.googlecode.com/'%VERSION
 # Configuration
 import config
 
@@ -28,11 +32,21 @@ else:
   from easylog import critical, error, warning, debug, info, setLogger, DEBUG
 #  setLogger(name='urssus', level=DEBUG)
 
-
+from feedupdater import updateOne
 
 # Some feeds put html in titles, which can't be shown in QStandardItems
 from html2text import html2text as h2t
 
+def detailToTitle(td):
+  '''Converts something like feedparser's title_detail into a 
+  nice text or html string.'''
+  # Title may be in plain title, but a title_detail is preferred
+  if td.type=='text/html':
+    title=h2t(td.value).strip().replace('\n', ' ')
+  else:
+    title=td.value.strip()    
+  return title
+
 def detailToAuthor(ad):
   '''Converts something like feedparser's author_detail into a 
   nice string describing the author'''
@@ -84,10 +98,17 @@ class Post(elixir.Entity):
   # Added in schema version 5
   fresh       = elixir.Field(elixir.Boolean, default=True)
 
+  decoTitle    = ''
+
   def __repr__(self):
-    if '<' in self.title:
-      return h2t(self.title).strip()
-    return unicode(self.title)
+    if not self.decoTitle:
+      self.decoTitle=h2t(self.title).strip().replace('\n', ' ')
+    return self.decoTitle
+    
+  def titleLink(self):
+    if self.link:
+      return '<a href="%s">%s</a>'%(self.link, unicode(self))
+    return self.title
 
 class Feed(elixir.Entity):
   elixir.using_options (tablename='feeds', inheritance='multi')
@@ -99,7 +120,7 @@ class Feed(elixir.Entity):
   children       = elixir.OneToMany('Feed', inverse='parent', order_by='position')
   parent         = elixir.ManyToOne('Feed')
   posts          = elixir.OneToMany('Post', order_by="-date", inverse='feed')
-  lastUpdated    = elixir.Field(elixir.DateTime, default=datetime(1970,1,1))
+  lastUpdated    = elixir.Field(elixir.DateTime, default=datetime.datetime(1970,1,1))
   loadFull       = elixir.Field(elixir.Boolean, default=False)
   # meaning of archiveType:
   # 0 = use default, 1 = keepall, 2 = use limitCount
@@ -118,13 +139,25 @@ class Feed(elixir.Entity):
   # Added in schema version 4
   is_open        = elixir.Field(elixir.Integer, default=0)
   curUnread      = -1
-
+  # Added in schema version 6
+  subtitle       = elixir.Field(elixir.Text, default='')
+  # Added in schema version 7
+  etag           = elixir.Field(elixir.Text, default='')
+  # Added in schema version 7
+  lastModified   = elixir.Field(elixir.DateTime, colname="last-modified", default=datetime.datetime(1970,1,1))
+  
   def __repr__(self):
-    c=self.unreadCount()
-    if c:
-      return self.text+'(%d)'%c
     return self.text
 
+  def titleLink(self):
+    if self.title:
+      title=self.title
+    else:
+      title=self.text
+    if self.htmlUrl:
+      return '<a href="%s">%s</a>'%(self.htmlUrl, title)
+    return title
+
   def markAsRead(self):
     if self.xmlUrl: # regular feed
       Post.table.update().where(Post.unread==True).where(Post.feed==self).values(unread=False).execute()
@@ -143,37 +176,45 @@ class Feed(elixir.Entity):
     # meaning of archiveType:
     # 0 = use default, 1 = keepall, 2 = use limitCount
     # 3 = use limitDays, 4 = no archiving
-    now=datetime.now()
+    now=datetime.datetime.now()
     if self.archiveType==0: # Default archive config
-      for post in self.posts:
-        if post.important: continue # Don't delete important stuff
-        # FIXME: makethis configurable
-	if (now-post.date).days>7: # Right now, keep for a week
-          post.deleted=True
-          post.unread=False
+      days=config.getValue('options', 'defaultExpiration', 7)
+      cutoff=now-datetime.timedelta(7, 0, 0)
+      Post.table.update().where(sql.and_(Post.important==False,  
+                                         Post.feed==self, 
+                                         Post.date<cutoff)).\
+                          values(deleted=True).execute()
+      elixir.session.flush()
     elif self.archiveType==1: #keepall
+      # Tested ;-)
       return
     elif self.archiveType==2: #limitCount
+      # FIXME: implement quicker!
+      # Doesn't seem to work
       for post in self.posts[self.limitCount:]:
         if post.important: continue # Don't delete important stuff
         post.deleted=True
+      elixir.session.flush()
     elif self.archiveType==3: #limitDays
-      for post in self.posts:
-        if post.important: continue # Don't delete important stuff
-        if (now-post.date).days>self.limitDays: # Right now, keep for a week
-          post.deleted=True
+      # Tested
+      cutoff=now-datetime.timedelta(self.limitDays, 0, 0)
+      Post.table.update().where(sql.and_(Post.important==False, 
+                                         Post.date<cutoff)).\
+                          values(deleted=True).execute()
+      elixir.session.flush()
     elif self.archiveType==4: #no archiving
-      for post in self.posts:
-        if post.important: continue # Don't delete important stuff
-        post.deleted=True
-      
-    Post.table.update().where(Post.deleted==True).values(unread=False).execute()
+      # Tested
+      Post.table.update().where(sql.and_(Post.important==False, 
+                                         Post.feed==self)).\
+                          values(deleted=True).execute()
+      elixir.session.flush()
+        
     if expunge:
       # Delete all posts with deleted==True, which are not fresh 
       # (are not in the last RSS/Atom we got)
       Post.table.delete().where(sql.and_(Post.deleted==True,Post.fresh==False,Post.feed==self)).execute()
 
-    elixir.session.flush()
+      elixir.session.flush()
       
     # Force recount
     self.curUnread=-1
@@ -191,9 +232,8 @@ class Feed(elixir.Entity):
   def allPostsQuery(self):
     '''Returns a query that should give you all the posts contained in this folder's children
     down to any level. Required for aggregate feeds, I'm afraid'''
-    af=self.allFeeds()
-    ored=sql.or_(*[Post.feed==f for f in af])
-    return Post.query().filter(ored)
+    af=[feed.id for feed in self.allFeeds()]
+    return Post.query.filter(Post.table._columns.get('feed_id').in_(af))
 
   def allPosts(self):
     '''This is used if you want all posts in this feed as well as in its childrens.
@@ -277,7 +317,7 @@ class Feed(elixir.Entity):
         if sib.xmlUrl:
           return sib
         else:
-          return sib.previousUnreadFeed()
+          return sib.lastChild().previousUnreadFeed()
     # Then see if our parent is the answer
     if self.parent and self.parent.unreadCount():
       if self.parent.xmlUrl:
@@ -335,31 +375,45 @@ class Feed(elixir.Entity):
     else:
       if self.curUnread==-1:
         info ("Forcing recount in %s", self.title)
-        self.curUnread=Post.query.filter(Post.feed==self)\
-                                 .filter(Post.unread==True)\
-                                 .filter(Post.deleted==False).count()
+        self.curUnread=Post.query.filter(Post.feed==self).filter(Post.deleted==False).filter(Post.unread==True).count()
     return self.curUnread
       
-  def updateFeedData(self):
+  def updateFeedData(self, parsedFeed):
+    # Fills blanks in feed data from parsedFeed
     # assumes this feed has a xmlUrl, fetches any missing data from it
     if not self.xmlUrl: # Nowhere to fetch data from
       return
-    info ( "Updating feed data from: %s", self.xmlUrl)
-    d=fp.parse(self.xmlUrl)
+      
+    # See if we have all data:
+    if self.htmlUrl and self.title and self.subtitle and self.description:
+      return
+    
+    d=parsedFeed
+    
     if not self.htmlUrl:
       if 'link' in d['feed']:
         self.htmlUrl=d['feed']['link']
       else:
         # This happens, for instance, on deleted blogger blogs (everyone's clever)
         self.htmlUrl=None
+        
     if not self.title:
-      if 'title' in d: 
-        self.title=d['feed']['title']
-        self.text=d['feed']['title']
+      if 'title_detail' in d['feed']: 
+        self.title=detailToTitle(d['feed']['title_detail'])
       else:
-        # This happens, for instance, on deleted blogger blogs (everyone's clever)
-        self.title=None
-        self.text=None
+        self.title=d['feed']['title']
+    else:
+      # This happens, for instance, on deleted blogger blogs (everyone's clever)
+      self.title=''
+
+    if not self.subtitle:
+      if 'subtitle_detail' in d['feed']: 
+        self.subtitle=detailToTitle(d['feed']['subtitle_detail'])
+      elif 'subtitle' in d['feed']:
+        self.subtitle=d['feed']['subtitle']
+      else:
+        self.subtitle=''
+
     if not self.description:
       if 'info' in d['feed']:
         self.description=d['feed']['info']
@@ -374,13 +428,36 @@ class Feed(elixir.Entity):
         pass #I am not going to care about errors here :-D
     elixir.session.flush()
     
-  def update(self):
+  def update(self, forced=False):
+    feedStatusQueue.put([0, self.id])
+    try:
+      self.real_update(forced)
+    except: # FIXME: reraise
+      pass
+    feedStatusQueue.put([1, self.id])
+    
+  def real_update(self, forced=False):
     if not self.xmlUrl: # Not a real feed
-      # FIXME: should update all children?
-      return
+      af=self.allFeeds()
+      for f in af:
+        f.update()
+
+    if self.lastModified:
+      mod=self.lastModified
+    else:
+      mod=datetime.datetime(1970, 1, 1)
+      
     if self.title:
       statusQueue.put(u"Updating: "+ self.title)
-    d=fp.parse(self.xmlUrl)
+    d=fp.parse(self.xmlUrl, etag=self.etag, modified=mod.timetuple())
+    
+    if d.status==304 and not forced: # No need to fetch
+      return
+    if d.status==301: # Permanent redirect
+      self.xmlUrl=d.href
+    if d.status==410: # Feed deleted. FIXME: tell the user and stop trying!
+      return
+
     posts=[]
     for post in d['entries']:
       try:
@@ -394,9 +471,9 @@ class Feed(elixir.Entity):
         else:
           dkey=None
         if dkey and post[dkey]:
-          date=datetime.fromtimestamp(time.mktime(post[dkey]))
+          date=datetime.datetime.fromtimestamp(time.mktime(post[dkey]))
         else:
-          date=datetime.now()
+          date=datetime.datetime.now()
         
         # So can the "unique ID for this entry"
         if 'id' in post:
@@ -440,12 +517,20 @@ class Feed(elixir.Entity):
           link=post['link']
         else:
             link=None
+            
+        # Titles may be in plain title, but a title_detail is preferred
+        if 'title_detail' in post:
+          title=detailToTitle(post['title_detail'])
+        else:
+          title=post['title']
+
+          
         # FIXME: if I use date to check here, I get duplicates on posts where I use
         # artificial date because it's not in the feed's entry.
         # If I don't I don't re-get updated posts.
-        p = Post.get_by(feed=self, title=post['title'],post_id=post[idkey])
+        p = Post.get_by(feed=self, title=title,post_id=post[idkey])
         if not p:
-          p=Post(feed=self, date=date, title=post['title'], 
+          p=Post(feed=self, date=date, title=title, 
                  post_id=post[idkey], content=content, 
                  author=author, link=link)
           if self.markRead:
@@ -453,85 +538,43 @@ class Feed(elixir.Entity):
           posts.append(p)
       except KeyError:
         debug( post )
-    self.lastUpdated=datetime.now()
+    self.lastUpdated=datetime.datetime.now()
+    self.updateFeedData(d)
+    if 'modified' in d:
+      self.lastModified=datetime.datetime(*d['modified'][:6])
+    if 'etag' in d:
+      self.etag=d['etag']
 
-    # Fix freshness
-    Post.table.update().values(fresh=False).execute()
-    for post in posts:
-      post.fresh=True
-    elixir.session.flush()
-      
-    # Queue a notification if needed
-    if posts and self.notify:
-      feedStatusQueue.put([3, self.id, len(posts)])
+    # Silly way to release the posts objects
+    # we don't need anymore
+    posts=[post.id for post in posts]
 
+    if posts:
+      with elixir.session.begin():
+        # Fix freshness
+        Post.table.update().where(sql.except_(Post.table.select(Post.id.in_(posts)))).values(fresh=False)
+        elixir.session.commit()
+
+      # Mark feed UI for updating
+      self.curUnread=-1
+      self.unreadCount()
+      feedStatusQueue.put([2, self.id])
+
+      # Queue a notification if needed
+      if self.notify:
+        feedStatusQueue.put([3, self.id, len(posts)])
+    
   def getQuery(self):
     if self.xmlUrl:
       return Post.query.filter(Post.feed==self)
     else:
       return self.allPostsQuery()
 
-  def nextPost(self, post, order, required=None, textFilter=''):
-    '''Returns next post in this feed after "post" or None'''
-    posts=self.getQuery()
-    if required:
-      posts=posts.filter(sql.or_(Post.id==post.id, required==True))
-    if textFilter:
-      posts=posts.filter(sql.or_(Post.id==post.id,Post.title.like('%%%s%%'%textFilter), Post.content.like('%%%s%%'%textFilter)))
-    posts=posts.order_by(order).all()
-    if posts:
-      ind=posts.index(post)
-      if ind+1<len(posts):
-        return posts[ind+1]
-    return None
-
-  def nextUnreadPost(self, post, order, required=None, textFilter=''):
-    '''Returns next unread post after "post" in this feed or None'''
-    posts=self.getQuery().filter(sql.or_(Post.unread==True, Post.id==post.id))
-    if required:
-      posts=posts.filter(sql.or_(Post.id==post.id, required==True))
-    if textFilter:
-      posts=posts.filter(sql.or_(Post.id==post.id,Post.title.like('%%%s%%'%textFilter), Post.content.like('%%%s%%'%textFilter)))
-    posts=posts.order_by(order).all()    
-    if posts:
-      ind=posts.index(post)
-      if ind+1<len(posts):
-        return posts[ind+1]
-    return None
-
-  def previousPost(self, post, order, required=None, textFilter=''):
-    '''Returns previous post in this feed or None'''
-    posts=self.getQuery()
-    if required:
-      posts=posts.filter(sql.or_(Post.id==post.id, required==True))
-    if textFilter:
-      posts=posts.filter(sql.or_(Post.id==post.id,Post.title.like('%%%s%%'%textFilter), Post.content.like('%%%s%%'%textFilter)))
-    posts=posts.order_by(order).all()
-    if posts:
-      ind=posts.index(post)
-      if ind>0:
-        return posts[ind-1]
-    return None
-
-  def previousUnreadPost(self, post, order, required=None, textFilter=''):
-    '''Returns previous post in this feed or None'''
-    posts=self.getQuery().filter(sql.or_(Post.unread==True, Post.id==post.id))
-    if required:
-      posts=posts.filter(sql.or_(Post.id==post.id, required==True))
-    if textFilter:
-      posts=posts.filter(sql.or_(Post.id==post.id,Post.title.like('%%%s%%'%textFilter), Post.content.like('%%%s%%'%textFilter)))
-    posts=posts.order_by(order).all()
-    if posts:
-      ind=posts.index(post)
-      if ind>0:
-        return posts[ind-1]
-    return None
-
 root_feed=None
 
 def initDB():
   global root_feed
-  REQUIRED_SCHEMA=5
+  REQUIRED_SCHEMA=9
   # FIXME: show what we are doing on the UI
   if not os.path.exists(database.dbfile): # Just create it
     os.system('urssus_upgrade_db')
@@ -543,10 +586,17 @@ def initDB():
     if curVer < REQUIRED_SCHEMA:
       info ("UPGRADING from %s to %s", curVer, REQUIRED_SCHEMA)
       os.system('urssus_upgrade_db')
+
+  import sqlite3
+  def connect():
+    conn = sqlite3.connect(database.dbfile)
+    conn.isolation_level = "IMMEDIATE"
+    return conn
+  engine = sql.create_engine(database.dbUrl, creator=connect) 
     
-  elixir.metadata.bind = database.dbUrl
+  elixir.metadata.bind = engine
   elixir.setup_all()
   elixir.session.flush()
   root_feed=Feed.get_by_or_init(parent=None)
+  root_feed.text='All Feeds'
   elixir.session.flush()
-  return root_feed
